<?php
/**
 * Configurator class.
 *
 * @category   apps
 * @package    Gateway Management
 * @subpackage libraries
 * @author     ADAMnetworks <support@adamnet.works>
 * @copyright  2015-2020 ADAMnetworks
 * @license    All rights reserved
 * @link       https://adamnet.works
 */

///////////////////////////////////////////////////////////////////////////////
// N A M E S P A C E
///////////////////////////////////////////////////////////////////////////////

namespace clearos\apps\gateway_management;

///////////////////////////////////////////////////////////////////////////////
// B O O T S T R A P
///////////////////////////////////////////////////////////////////////////////

$bootstrap = getenv('CLEAROS_BOOTSTRAP') ? getenv('CLEAROS_BOOTSTRAP') : '/usr/clearos/framework/shared';
require_once $bootstrap . '/bootstrap.php';

///////////////////////////////////////////////////////////////////////////////
// D E P E N D E N C I E S
///////////////////////////////////////////////////////////////////////////////

// Classes
//--------

use \clearos\apps\base\Configuration_File;
use \clearos\apps\base\Engine;
use \clearos\apps\base\File;

clearos_load_library('base/Configuration_File');
clearos_load_library('base/Engine');
clearos_load_library('base/File');

// Exceptions
//-----------
use \clearos\apps\base\Engine_Exception;
use \clearos\apps\base\Not_Found_Exception;
use \clearos\apps\base\Validation_Exception as Validation_Exception;

clearos_load_library('base/Engine_Exception');


///////////////////////////////////////////////////////////////////////////////
// C L A S S
///////////////////////////////////////////////////////////////////////////////

/**
 * Configurator class.
 *
 * @category   apps
 * @package    Gateway Management
 * @subpackage libraries
 * @author     ADAMnetworks
 * @copyright  2014-2019 ADAMnetworks
 * @license    All rights reserved
 * @link       https://adamnet.works/
 */

class Configurator extends Engine
{

    ///////////////////////////////////////////////////////////////////////////////
    // V A R I A B L E S
    ///////////////////////////////////////////////////////////////////////////////

    protected $data = NULL;
    protected $restore_path = NULL;
    protected $path;
    protected $is_loaded = FALSE;

    ///////////////////////////////////////////////////////////////////////////////
    // M E T H O D S
    ///////////////////////////////////////////////////////////////////////////////

    /**
     * Configuration constructor.
     *
     * @param $path          The path to load.
     * @param $restore_path  The path to log initial values.
     */
    public function __construct($path, $restore_path=NULL)
    {
        clearos_profile(__METHOD__, __LINE__);

        $this->path = $path;
        $this->restore_path = $restore_path;
    }

    /**
     * Add a comment.
     *
     * @param string $comment text
     *
     * @throws Engine_Exception
     */
    public function comment($comment)
    {
        clearos_profile(__METHOD__, __LINE__);

        try {
            // Update the setting.
            $file = new File($this->path, TRUE);
            $file->add_lines("# $comment\n");

        } catch (Exception $e) {
            throw new Engine_Exception(clearos_exception_message($e), CLEAROS_ERROR);
        }
    }

    /**
     * Get a value.
     *
     * @return string/boolean  String value, true if key-only, false if not there.
     *
     * @throws Engine_Exception
     */
    public function get($key)
    {
        clearos_profile(__METHOD__, __LINE__);

        $config = $this->_load();

        // New lines can mess up the data.
        if ($key === '') {
            return FALSE;
        }

        // Not set.
        if (!isset($config[$key])) {
            return FALSE;
        }

        // Key-only value.
        if ($config[$key] === '') {
            return TRUE;
        }

        return $config[$key];
    }

    /**
     * Get all values.
     *
     * @return array  All key/value pairs, with value following the format of get().
     *
     * @throws Engine_Exception
     */
    public function get_all()
    {
        clearos_profile(__METHOD__, __LINE__);

        $this->_load();

        // Loop over keys to get the data in the proper format.
        $results = array();
        foreach($this->data as $key => $value) {
            // New lines can mess up the data.
            if ($key === '') {
                continue;
            }

            $results[$key] = $this->get($key);
        }

        return $results;
    }


    /**
     * Set a value.
     *
     * @param string    $key  key name
     * @param string/boolean  $value  value for the key. True for key only, false to delete.
     *
     * @return  bool changed
     * @throws Engine_Exception
     */
    public function set($key, $value)
    {
        clearos_profile(__METHOD__, __LINE__);

        if ($key == '') {
            throw new Engine_Exception('Invalid key provided.', CLEAROS_ERROR);
        }

        if ($this->verify($key, $value)) {
            return FALSE;
        }

        try {
            // If needed, store the current value.
            $current = $this->get($key);
            if ($this->restore_path && $current) {
                $changes = new Configurator($this->restore_path);
                $changes->set($key, $current);
            }

            // Update the setting.
            $file = new File($this->path, TRUE);

            if ($value === FALSE) {
                $file->delete_lines("/^$key\s*(=\s*)?/");
            } else {
                if ($value === TRUE) {
                    $line = "$key\n";
                } else {
                    $line = "$key=$value\n";
                }

                $matched = $file->replace_lines("/^$key\s*(=\s*)?/", $line);
                if (!$matched) {
                    $file->add_lines($line);
                }
            }

        } catch (Exception $e) {
            throw new Engine_Exception(clearos_exception_message($e), CLEAROS_ERROR);
        }

        $this->is_loaded = FALSE;
    }

    /**
     * Verify a parameter is correct.
     *
     * @param string $key   key name
     * @param string/boolean $value value for the key
     *
     * @return bool
     * @throws Engine_Exception
     */
    public function verify($key, $value)
    {
        clearos_profile(__METHOD__, __LINE__);

        return ($this->get($key) === $value);
    }

    /**
     * Reset parameters to pre-install state.
     *
     * @return boolean changes made
     * @throws Engine_Exception
     */
    public function restore()
    {
        clearos_profile(__METHOD__, __LINE__);

        if (!$this->restore_path) {
            throw new Engine_Exception("Restore is not available for \"$this->path\". Try delete() instead.", CLEAROS_ERROR);
        }

        // Load the original config.
        $original = new Configurator($this->restore_path);

        // Change the values.
        $changed = 0;
        foreach ($original->get_all() as $key => $value) {
            $changed += $this->set($key, $value);
        }

        // Remove the changes file.
        $file = new File($this->restore_path);
        try {
            $file->delete();
        } catch (Engine_Exception $e) {
            // File may not exist, so we don't need to do anything.
        }

        return (bool)$changed;
    }

    /**
     * Delete the config file..
     *
     * @return boolean changes made
     * @throws Engine_Exception
     */
    public function delete()
    {
        if ($this->restore_path) {
            throw new Engine_Exception("Delete is not available for \"$this->path\". Try restore() instead.", CLEAROS_ERROR);
        }

        // Remove the changes file.
        $file = new File($this->path);
        if ($file->exists()) {
            $file->delete();
        }

        $this->data = NULL;
    }

    /**
     * Get a list of changed keys.
     *
     * @param array $original   Original values
     *
     * @return array changed keys
     */
    public function get_changed($original)
    {
        clearos_profile(__METHOD__, __LINE__);

        $changes = [];

        // Check all the items.
        $current = $this->get_all();
        $merged = array_merge($original, $current);
        foreach ($merged as $key => $value) {

            // See if value's unchanged.
            if (isset($original[$key]) && $original[$key] === $this->get($key)) {
                continue;
            }

            $changes[] = $key;
        }

        return $changes;
    }

    /**
     * Reset the config file to empty state.
     *
     * @return boolean changes made
     * @throws Engine_Exception
     */
    public function clear()
    {
        $this->delete();
        $this->_create();
    }

    ///////////////////////////////////////////////////////////////////////////////
    // P R I V A T E   M E T H O D S
    ///////////////////////////////////////////////////////////////////////////////

    /**
     * Create configuration file.
     *
     * @return array
     *
     * @throws Engine_Exception
     */
    protected function _create()
    {
        clearos_profile(__METHOD__, __LINE__);

        $file = new Configuration_File($this->path, 'match', "/([^=]*)\s*=?\s*(.*)?/");

        if (!$file->exists()) {
            $file->create('webconfig', 'webconfig', '0644');
        }

        $this->data = $file->load();
    }

    /**
     * Loads configuration file.
     *
     * @return array
     *
     * @throws Engine_Exception
     */
    protected function _load()
    {
        clearos_profile(__METHOD__, __LINE__);

        if (!$this->is_loaded) {

            try {
                $file = new Configuration_File($this->path, 'match', "/([^=]*)\s*=?\s*(.*)?/");

                if (!$file->exists()) {
                    $file->create('webconfig', 'webconfig', '0644');
                }

                $this->data = $file->load();
            } catch (Exception $e) {
                throw new Engine_Exception(clearos_exception_message($e), CLEAROS_ERROR);
            }
        }

        $this->is_loaded = TRUE;

        return $this->data;
    }

}
