<?php
/**
 * Firewaller class.
 *
 * @category   apps
 * @package    Gateway Management
 * @subpackage libraries
 * @author     ADAMnetworks <support@adamnet.works>
 * @copyright  2015-2019 ADAMnetworks
 * @license    All rights reserved
 * @link       https://adamnet.works
 */

///////////////////////////////////////////////////////////////////////////////
// N A M E S P A C E
///////////////////////////////////////////////////////////////////////////////

namespace clearos\apps\gateway_management;

///////////////////////////////////////////////////////////////////////////////
// B O O T S T R A P
///////////////////////////////////////////////////////////////////////////////

$bootstrap = getenv('CLEAROS_BOOTSTRAP') ? getenv('CLEAROS_BOOTSTRAP') : '/usr/clearos/framework/shared';
require_once $bootstrap . '/bootstrap.php';

///////////////////////////////////////////////////////////////////////////////
// D E P E N D E N C I E S
///////////////////////////////////////////////////////////////////////////////

// Classes
//--------

use \clearos\apps\gateway_management\Network;
use \clearos\apps\gateway_management\Util;

use \clearos\apps\base\Engine;
use \clearos\apps\base\File;

clearos_load_library('gateway_management/Network');
clearos_load_library('gateway_management/Util');

clearos_load_library('base/Engine');
clearos_load_library('base/File');

// Exceptions
//-----------
use \clearos\apps\base\Engine_Exception;
use \clearos\apps\base\Not_Found_Exception;
use \clearos\apps\base\Validation_Exception as Validation_Exception;

clearos_load_library('base/Engine_Exception');


///////////////////////////////////////////////////////////////////////////////
// C L A S S
///////////////////////////////////////////////////////////////////////////////

class Firewall extends Engine
{
    ///////////////////////////////////////////////////////////////////////////////
    // C O N S T A N T S
    ///////////////////////////////////////////////////////////////////////////////

    // Config
    const FIREWALL_RULES = '/etc/clearos/firewall.d/10-adamnet';
    const SYSCTL_CONFIG  = '/usr/lib/sysctl.d/10-adamnet.conf';  # Run after 00-system

    // Firewall tables.
    const TABLE_DEFAULT = 'ADAMNET';
    const TABLE_BLOCK = 'ADAMNETBLOCK';
    const TABLE_IPE = 'ADAMNETIPE';
    const TABLE_REJECT = 'ADAMNETREJECT';

    // IPSets.
    const IP_DASHBOARD4 = 'an_dash4';
    const IP_WA4 = 'an_wa4';
    const IP_WA6 = 'an_wa6';

    ///////////////////////////////////////////////////////////////////////////////
    // V A R I A B L E S
    ///////////////////////////////////////////////////////////////////////////////

    protected $dashboard_ips = [
        '34.98.73.57',     // adamnet.works
        '34.120.175.252',  // dashboard.adamnet.works, dashboard.contentfilter.net
    ];

    ///////////////////////////////////////////////////////////////////////////////
    // M E T H O D S
    ///////////////////////////////////////////////////////////////////////////////

    /**
     * Firewall constructor.
     */
    public function __construct()
    {
        clearos_profile(__METHOD__, __LINE__);
    }

    /**
     * Update the firewall.
     *
     * @param bool $enabled Service enabled
     * @param array $config App config options
     *
     * @return bool changed
     */
    public function update($enabled, $config)
    {
        clearos_profile(__METHOD__, __LINE__);

        $fw = array();

        // Set the paths.
        $util = new Util();
        $fw[] = "cat="      . $util->get_path('cat');
        $fw[] = "ebtables=" . $util->get_path('ebtables');
        $fw[] = "kill="     . $util->get_path('kill');

        // Get the rules.
        $fw[] = ($enabled) ? $this->_get_firewall_enabled($config) : $this->_get_firewall_disabled();

        // make sure file exists
        $file = new File(self::FIREWALL_RULES, TRUE);
        if (!$file->exists()) {
            $file->create('webconfig', 'webconfig', '0755');
        }

        // update contents
        $current_contents = $file->get_contents();
        $current_contents = trim($current_contents);

        $lines = implode("\n", $fw);
        if ($current_contents != $lines) {
            $file->delete_lines('/.*/');
            $file->add_lines($lines);
            return TRUE;
        }

        // no changes
        return FALSE;
    }

    ///////////////////////////////////////////////////////////////////////////////
    // P R I V A T E   M E T H O D S
    ///////////////////////////////////////////////////////////////////////////////

    /**
     * Get firewall rules for when this is disabled.
     *
     * @return string rules
     */
    private function _get_firewall_disabled()
    {
        clearos_profile(__METHOD__, __LINE__);

        return $this->_get_ebtables_disabled();
    }

    /**
     * Get firewall rules for when this is enabled.
     *
     * @param array $config Config options
     *
     * @return string rules
     */
    private function _get_firewall_enabled($config)
    {
        clearos_profile(__METHOD__, __LINE__);

        $fw = array();
        $eb = array();

        $DEFAULT = self::TABLE_DEFAULT;
        $BLOCK = self::TABLE_BLOCK;
        $IPE = self::TABLE_IPE;
        $REJECT = self::TABLE_REJECT;
        $wa4 = self::IP_WA4;
        $dash4 = self::IP_DASHBOARD4;

        $network = new Network();

        // IPv4 rules (zero-length test for ClearOS 6.6 compatibility)
        $fw[] = 'if [ "$FW_PROTO" == "ipv4" ] || [ -z "$FW_PROTO" ]; then';

        // Set up ipset
        $fw[] = "  ipset create -exist $wa4 hash:ip family ipv4";

        // Setup wan access (AKA: No Internet) and IPE (AKA: DTTS)
        $fw[] .= "  \$IPTABLES -N $IPE";
        $fw[] .= "  \$IPTABLES -N $REJECT";

        // Reject rules.
        // Exceptions for dashboard IPs come first.
        $fw[] = "  ipset create -exist $dash4 hash:ip family ipv4";
        foreach ($this->dashboard_ips as $dash_ip) {
            $fw[] = "  ipset add -exist $dash4 $dash_ip";
        }
        $fw[] = "  \$IPTABLES -A $REJECT -p tcp -m set --match-set $dash4 dst --dport 443 -j RETURN";
        $fw[] = "  \$IPTABLES -A $REJECT -p tcp -j REJECT --reject-with tcp-reset";
        $fw[] = "  \$IPTABLES -A $REJECT -j REJECT --reject-with icmp-admin-prohibited";  // see https://tools.ietf.org/html/rfc1812#section-5.2.7.1

        // Block WAN rule first.
        $fw[] = "  \$IPTABLES -I FORWARD 1 -m set --match-set $wa4 src -j $REJECT";

        // See if we have an established connections rule.
        // Disabling Internet access cuts these off.
        // Note: Uses "tun+" instead of "ESTABLISHED" to work with OpenVPN.
        //       We need the last "tun+" (if more than one).
        //       Before this rule existed, ClearCenter told people to add
        //       an additional tun+ rule higher up as a workaround.
        $fw[] = '  rule=$($IPTABLES -t filter -L FORWARD -nv --line-numbers | grep tun+ | tail -n 1 | cut -d" " -f1)';
        $fw[] = '  if [ -z "$rule" ]; then';
        // If there is no rule for established traffic, use rule number 1.
        // This is needed for bridge mode.
        $fw[] = "  rule=1";
        $fw[] = "  fi";

        // Create jump to IPE rules (after established connections).
        $fw[] = "  ((rule = rule+1))";
        $fw[] = "  \$IPTABLES -t filter -I FORWARD \$rule -j $IPE";

        // Block page IP
        $http_addr = $config->get('http-listen-address');
        $proxy_addr = $config->get('http-proxy-subnet');
        $http_port = $config->get('http-listen-port');

        $fw[] = "  \$IPTABLES -t filter -N $BLOCK";
        $fw[] = "  \$IPTABLES -t filter -I INPUT -d $http_addr -j $BLOCK";

        $fw[] = "  \$IPTABLES -t filter -A $BLOCK -p tcp -m multiport --dport 80,6373 -j ACCEPT";
        $fw[] = "  \$IPTABLES -t filter -A $BLOCK -p tcp --dport 443 -j REJECT --reject-with=tcp-reset";
        $fw[] = "  \$IPTABLES -t filter -A $BLOCK -p tcp -j REJECT";
        $fw[] = "  \$IPTABLES -t filter -A $BLOCK -p udp -j REJECT";

        // For DNS hijack
        $fw[] = "  \$IPTABLES -t nat -N $DEFAULT";
        $fw[] = "  \$IPTABLES -t nat -I PREROUTING -j $DEFAULT";
        $fw[] = "  \$IPTABLES -t nat -I $DEFAULT 1 -p tcp -d $http_addr --dport 80 -j DNAT --to $http_addr:$http_port";

        // Per-interface rules
        $lan_ifs = $network->get_interfaces(Network::FILTER_LAN);
        foreach ($lan_ifs as $lan_if => $lan_data) {
            if (isset($lan_data['ifcfg']['ipaddr'])) {
                $lan_ip = $lan_data['ifcfg']['ipaddr'];

                // Hijack any DNS queries not directed to this IP
                $fw[] = "  \$IPTABLES -t nat -I $DEFAULT 1 -p tcp -i $lan_if ! -d $lan_ip --dport 53 -j DNAT --to $lan_ip:53";
                $fw[] = "  \$IPTABLES -t nat -I $DEFAULT 1 -p udp -i $lan_if ! -d $lan_ip --dport 53 -j DNAT --to $lan_ip:53";

                if ($lan_data['type'] === 'Bridge' && isset($lan_data['ifcfg']['netmask'])) {
                    // We need to accept any non-DNS traffic destined for this network.
                    // This allows other connections to the gateway to work,
                    // and it allows responses incoming traffic with DTTS enabled.
                    $netmask = $lan_data['ifcfg']['netmask'];
                    $fw[] = "  \$IPTABLES -I FORWARD 1 -d $lan_ip/$netmask -j ACCEPT";
                }
            }

            if ($lan_data['type'] === 'Bridge') {
                // Accept the mytools and block page traffic
                $eb[] = "--logical-in $lan_if -p ipv4 --ip-dst $http_addr -j redirect";

                // Accept subnet proxy traffic
                if ($proxy_addr) {
                    $eb[] = "--logical-in $lan_if -p ipv4 --ip-dst $proxy_addr -j redirect";
                }
            }
        }

        // Using a DNS server on the LAN disables hijacking for that IP.
        $internal_ips = $network->get_internal_dns_ips();
        foreach ($internal_ips as $ip) {
            $fw[] = "  \$IPTABLES -t nat -I $DEFAULT -p tcp --dport 53 -s $ip -j ACCEPT";
            $fw[] = "  \$IPTABLES -t nat -I $DEFAULT -p udp --dport 53 -s $ip -j ACCEPT";
        }

        // IPv6 rules (for future use)
        // $fw .= 'elif [ "$FW_PROTO" == "ipv6" ]; then' . "\n";

        // None yet
        $fw[] = "fi";

        // Make sure we have the required kernel module for DTTS.
        $fw[] = "modprobe nfnetlink_queue";

        if ($network->get_mode() == 'trustedgateway') {
            // Bridging requires a kernel module to send traffic through iptables
            $fw[] = "modprobe br_netfilter";

            // Configure ebtables.
            // Create the chain
            $fw[] = "\$ebtables -t nat -N $DEFAULT 2> /dev/null";
            // Jump to chain
            $fw[] = "\$ebtables -t nat -L PREROUTING | grep -- '-j $DEFAULT' &> /dev/null || \$ebtables -t nat -I PREROUTING -j $DEFAULT";

            // Compare the current ebtables rules to the desired rules
            $eb_string = implode("\n", $eb);
            $fw[] = "EB_RULES_1='$eb_string'";
            $fw[] = "EB_RULES_2=\$(\$ebtables -t nat -L $DEFAULT 2> /dev/null | tail -n +4)";
            $fw[] = 'if [ "$EB_RULES_1" != "$EB_RULES_2" ]; then';

            // This isn't a match, so dump and recreate
            $fw[] = "  \$ebtables -t nat -F $DEFAULT";
            foreach ($eb as $rule) {
                $fw[] = "  \$ebtables -t nat -A $DEFAULT $rule";
            }
            $fw[] = "fi";
        } else {
            // Disable bridging rules
            $fw[] = $this->_get_ebtables_disabled();
        }

        // The firewall rules will sometimes be wiped and reloaded.
        // Tell the manager to rebuild its rules.
        // Note: This should be the last command.
        // If the file does not exist, this will fail and no more rules will be inserted.
        $fw[] = "\$kill -SIGUSR1 \$(\$cat /var/run/anmgr.pid)";

        // Return the lines
        $fw = implode("\n", $fw);
        return $fw;
    }

    /**
     * Get ebtables rules for when this is disabled.
     *
     * @return string rules
     */
    private function _get_ebtables_disabled()
    {
        clearos_profile(__METHOD__, __LINE__);

        // Note: "service firewall restart" shows error if this fails,
        // so we call true to set a success status code.
        $fw = array();

        $DEFAULT = self::TABLE_DEFAULT;

        $fw[] = '{';
        $fw[] = "  \$ebtables -t nat -D PREROUTING -j $DEFAULT";
        $fw[] = "  \$ebtables -t nat -X $DEFAULT";
        $fw[] = '} || true';

        return implode("\n", $fw);
    }

}
