<?php
/**
 * Network class.
 *
 * @category   apps
 * @package    Gateway Management
 * @subpackage libraries
 * @author     ADAMnetworks <support@adamnet.works>
 * @copyright  2015-2019 ADAMnetworks
 * @license    All rights reserved
 * @link       https://adamnet.works
 */

///////////////////////////////////////////////////////////////////////////////
// N A M E S P A C E
///////////////////////////////////////////////////////////////////////////////

namespace clearos\apps\gateway_management;

///////////////////////////////////////////////////////////////////////////////
// B O O T S T R A P
///////////////////////////////////////////////////////////////////////////////

$bootstrap = getenv('CLEAROS_BOOTSTRAP') ? getenv('CLEAROS_BOOTSTRAP') : '/usr/clearos/framework/shared';
require_once $bootstrap . '/bootstrap.php';

///////////////////////////////////////////////////////////////////////////////
// D E P E N D E N C I E S
///////////////////////////////////////////////////////////////////////////////

// Classes
//--------

use \clearos\apps\gateway_management\Util;

use \clearos\apps\base\Engine;
use \clearos\apps\base\Shell;
use \clearos\apps\network\Iface_Manager;
use \clearos\apps\network\Iface;
use \clearos\apps\network\Network as ClearNetwork;
use \clearos\apps\network\Network_Utils;
use \clearos\apps\network\Resolver;

clearos_load_library('gateway_management/Util');

clearos_load_library('base/Engine');
clearos_load_library('base/Shell');
clearos_load_library('network/Iface_Manager');
clearos_load_library('network/Iface');
clearos_load_library('network/Network_Utils');
clearos_load_library('network/Network');
clearos_load_library('network/Resolver');

// Exceptions
//-----------
use \clearos\apps\base\Engine_Exception;
use \clearos\apps\base\Not_Found_Exception;
use \clearos\apps\base\Validation_Exception as Validation_Exception;

clearos_load_library('base/Engine_Exception');


///////////////////////////////////////////////////////////////////////////////
// C L A S S
///////////////////////////////////////////////////////////////////////////////

class Network extends Engine
{

    ///////////////////////////////////////////////////////////////////////////////
    // C O N S T A N T S
    ///////////////////////////////////////////////////////////////////////////////

    // Interface filters
    const FILTER_EXTERNAL = 'external';
    const FILTER_LAN = 'lan';
    const FILTER_PHYSICAL = 'physical';

    ///////////////////////////////////////////////////////////////////////////////
    // V A R I A B L E S
    ///////////////////////////////////////////////////////////////////////////////

    protected $mode = NULL;

    ///////////////////////////////////////////////////////////////////////////////
    // M E T H O D S
    ///////////////////////////////////////////////////////////////////////////////

    /**
     * Network constructor.
     */
    public function __construct()
    {
        clearos_profile(__METHOD__, __LINE__);

        $this->get_mode();
    }

    /**
     * Get network mode.
     *
     * @return string mode
     */
    public function get_mode() {
        if (!$this->mode) {
            $network = new ClearNetwork();
            $this->mode = $network->get_mode();
        }
        return $this->mode;
    }

    /**
     * Get all network interfaces.
     *
     * @param string $filter Network filter
     *
     * @return array interfaces
     */
    public function get_interfaces($filter=NULL)
    {
        clearos_profile(__METHOD__, __LINE__);

        $options['filter_virtual'] = FALSE;
        $iface_manager = new Iface_Manager();
        $interfaces = $iface_manager->get_interface_details($options);

        // Filter interfaces.
        $filtered = array();
        foreach ($interfaces as $name => $iface) {
            if ($filter !== NULL && !$this->_interface_is_type($iface, $filter)) {
                continue;
            }

            $filtered[$name] = $iface;
        }

        return $filtered;
    }

    /**
     * Get specific network interfaces.
     *
     * @param string $iface interface
     *
     * @return array interfaces
     * @throws Not_Found_Exception
     */
    public function get_interface($iface)
    {
        clearos_profile(__METHOD__, __LINE__);

        $interfaces = $this->get_interfaces();

        if (!$interfaces[$iface]) {
            throw new Not_Found_Exception("Unknown interface $iface.", CLEAROS_ERROR);
        }

        return $interfaces[$iface];
    }

    /**
     * Check if an IP is internal.
     *
     * @return bool
     */
    public function is_internal_ip($ip)
    {
        clearos_profile(__METHOD__, __LINE__);

        $util = new Util();
        $ip_path = $util->get_path('ip');
        $grep_path = $util->get_path('grep');

        // Check the IP
        $shell = new Shell();
        $shell->execute($ip_path, "route get $ip | $grep_path -o -P '(?<=dev )[^\s]+'");

        $result = $shell->get_first_output_line();

        if ($result == 'lo') {
            // local interface
            return TRUE;
        }

        if ($this->get_mode() == 'trustedgateway') {
            // TODO: figure out how to detect internal IP in bridge mode
            // For now, just report false.
            return FALSE;
        }

        $iface = $this->get_interface($result);

        return ($iface['role'] !== 'EXTIF');
    }

    /**
     * Make sure an IP address is available.
     *
     * @param string $ip IP address to check
     *
     * @return bool
     */
    public function is_ip_in_use($ip)
    {
        clearos_profile(__METHOD__, __LINE__);

        $util = new Util();
        $ping_path = $util->get_path('ping');
        $shell = new Shell();

        // initialize status
        $status = TRUE;

        // ping the IP
        try {
            // 0 means response received, anything else is failure.
            $exit_code = $shell->execute($ping_path, "-c 1 $ip");
            $status = !$exit_code;
        } catch (Engine_Exception $e) {
            // no response, so probably not in use
            $status = FALSE;
        }

        // For unresponsive internal IP addresses, check ARP as well.
        if (!$status && $this->is_internal_ip($ip)) {
            try {
                $shell->execute('/sbin/arp', "-a -n | grep '$ip) at <incomplete>'");
                $status = FALSE;
            } catch (Engine_Exception $e) {
                // not critical
            }
        }

        return $status;
    }

    /**
     * Get internal DNS IPs.
     *
     * @return array
     */
    public function get_internal_dns_ips()
    {
        clearos_profile(__METHOD__, __LINE__);

        $internal_dns = array();

        // Get DNS.
        $resolver = new Resolver();
        $resolvers = $resolver->get_nameservers();
        foreach ($resolvers as $ip) {
            if ($this->is_internal_ip($ip)) {
                $internal_dns[] = $ip;
            }
        }

        return $internal_dns;
    }

    /**
     * Get the network of every interface.
     *
     * @return array all networks
     */
    public function interface_networks()
    {
        clearos_profile(__METHOD__, __LINE__);

        $networks = array();

        // validate IP is not on any networks
        $ifaces = $this->get_interfaces();
        foreach ($ifaces as $iface => $info) {
            // skip any inactive interfaces
            if (!$info['address'])
                continue;

            $iface_details = new Iface($iface);
            $network = $iface_details->get_network();

            // external interfaces have an empty string as network
            if (empty($network)) {
                continue;
            }

            $networks[$iface] = $network;
        }

        return $networks;
    }

    /**
     * Check if the given IP is on any networks.
     *
     * @param string $validate_ip IP address to validate
     *
     * @return bool true if on a network
     */
    public function ip_on_networks($validate_ip)
    {
        clearos_profile(__METHOD__, __LINE__);

        $network_utils = new Network_Utils();

        // validate IP is not on any networks
        $networks = $this->interface_networks();
        foreach ($networks as $network_name => $network) {
            if ($network_utils->ip_on_network($validate_ip, $network)) {
                return TRUE;
            }
        }
    }

    /**
     * Check if the given network overlaps with any of the interface networks.
     *
     * @param string $validate_network Network to validate
     *
     * @return string/null overlapping interface name
     */
    public function network_overlaps_interfaces($validate_network)
    {
        clearos_profile(__METHOD__, __LINE__);

        $network_utils = new Network_Utils();

        $validate_info = $this->get_network_info($validate_network);

        // validate IP is not on any networks
        $networks = $this->interface_networks();
        foreach ($networks as $network_name => $network) {
            // Get broadcast and network address from both
            $net_info = $this->get_network_info($network);

            if (
                $network_utils->ip_on_network($validate_info['address'], $network) ||
                $network_utils->ip_on_network($validate_info['broadcast'], $network) ||
                $network_utils->ip_on_network($net_info['address'], $validate_network) ||
                $network_utils->ip_on_network($net_info['broadcast'], $validate_network)
            ) {
                return $network_name;
            }
        }
    }

    /**
     * Get IP and netmask for a network.
     *
     * @param string $network Network to process
     *
     * @return array includes address, broadcast, ip, netmask, prefix
     */
    public function get_network_info($network)
    {
        clearos_profile(__METHOD__, __LINE__);

        $network_utils = new Network_Utils();

        list($ip, $prefix) = explode('/', $network);
        $netmask = $network_utils->get_netmask($prefix);

        $start = $network_utils->get_network_address($ip, $netmask);
        $end = $network_utils->get_broadcast_address($ip, $netmask);

        return array('address' => $start, 'broadcast' => $end, 'ip' => $ip, 'netmask' => $netmask, 'prefix' => $prefix);
    }


    ///////////////////////////////////////////////////////////////////////////////
    // P R I V A T E   M E T H O D S
    ///////////////////////////////////////////////////////////////////////////////

    /**
     * Determine if interface is specified type.
     *
     * @param string $iface interface
     * @param string $type Network filter
     *
     * @return bool valid
     * @throws Not_Found_Exception
     */
    private function _interface_is_type($iface, $type) {
        switch ($type) {
            case self::FILTER_EXTERNAL:
                return $this->_is_external_interface($iface);

            case self::FILTER_LAN:
                return $this->_is_lan_interface($iface);

            case self::FILTER_PHYSICAL:
                return $this->_is_physical_interface($iface);

            default:
                throw new Not_Found_Exception("Unknown interface $iface.", CLEAROS_ERROR);
        }
    }

    /**
     * Is this an external interface?
     *
     * @param string $iface interface
     *
     * @return bool is external
     */
    public function _is_external_interface($iface)
    {
        clearos_profile(__METHOD__, __LINE__);

        return ($iface['configured'] && $iface['role'] === 'EXTIF');
    }

    /**
     * Is this a LAN interface.
     *
     * @param string $iface interface
     *
     * @return bool is lan
     */
    public function _is_lan_interface($iface)
    {

        if (!$this->_is_physical_interface($iface)) {
            return FALSE;
        }

        // Have to have an IP address
        if (empty($iface['address'])) {
            return FALSE;
        }

        // Detect transparent bridges.
        if ($this->mode == 'trustedgateway' && $iface['type'] == Iface::TYPE_BRIDGED) {
            return TRUE;
        }

        if ($iface['role'] == 'EXTIF') {
            if ($this->mode == 'standalone' || $this->mode == 'trustedstandalone' || $this->mode == 'trustedgateway') {
                return TRUE;
            }
        } else {
            if ($this->mode == 'gateway') {
                return TRUE;
            }
        }

        return FALSE;
    }

    /**
     * Is this a physical interface?
     *
     * @param string $iface interface
     *
     * @return bool is physical
     */
    public function _is_physical_interface($iface)
    {
        clearos_profile(__METHOD__, __LINE__);

        return ($iface['configured'] && $iface['type'] !== Iface::TYPE_VIRTUAL);
    }

}
