<?php
/**
 * Service class.
 *
 * @category   apps
 * @package    Gateway Management
 * @subpackage libraries
 * @author     ADAMnetworks <support@adamnet.works>
 * @copyright  2015-2020 ADAMnetworks
 * @license    All rights reserved
 * @link       https://adamnet.works
 */

///////////////////////////////////////////////////////////////////////////////
// N A M E S P A C E
///////////////////////////////////////////////////////////////////////////////

namespace clearos\apps\gateway_management;

///////////////////////////////////////////////////////////////////////////////
// B O O T S T R A P
///////////////////////////////////////////////////////////////////////////////

$bootstrap = getenv('CLEAROS_BOOTSTRAP') ? getenv('CLEAROS_BOOTSTRAP') : '/usr/clearos/framework/shared';
require_once $bootstrap . '/bootstrap.php';

///////////////////////////////////////////////////////////////////////////////
// T R A N S L A T I O N S
///////////////////////////////////////////////////////////////////////////////

clearos_load_language('base');
clearos_load_language('gateway_management');

///////////////////////////////////////////////////////////////////////////////
// D E P E N D E N C I E S
///////////////////////////////////////////////////////////////////////////////

// Classes
//--------

use \clearos\apps\gateway_management\Configurator;
use \clearos\apps\gateway_management\Firewall;
use \clearos\apps\gateway_management\Network;

use \clearos\apps\base\Daemon;
use \clearos\apps\base\File;
use \clearos\apps\base\Folder;
use \clearos\apps\base\Shell;
use \clearos\apps\clearcenter\Rest;
use \clearos\apps\dhcp\Dnsmasq;
use \clearos\apps\marketplace\Marketplace as Marketplace;
use \clearos\apps\network\Hosts;
use \clearos\apps\network\Iface;
use \clearos\apps\network\Network_Utils;
use \clearos\apps\network\Resolver;
use \clearos\apps\network\Routes;
use \clearos\apps\tasks\Cron;

clearos_load_library('gateway_management/Configurator');
clearos_load_library('gateway_management/Firewall');
clearos_load_library('gateway_management/Network');

clearos_load_library('base/Daemon');
clearos_load_library('base/File');
clearos_load_library('base/Folder');
clearos_load_library('base/Shell');
clearos_load_library('clearcenter/Rest');
clearos_load_library('dhcp/Dnsmasq');
clearos_load_library('marketplace/Marketplace');
clearos_load_library('network/Hosts');
clearos_load_library('network/Iface');
clearos_load_library('network/Network_Utils');
clearos_load_library('network/Resolver');
clearos_load_library('network/Routes');
clearos_load_library('tasks/Cron');

// Exceptions
//-----------
use \clearos\apps\base\Engine_Exception;
use \clearos\apps\base\Not_Found_Exception;
use \clearos\apps\base\Validation_Exception as Validation_Exception;

clearos_load_library('base/Engine_Exception');


///////////////////////////////////////////////////////////////////////////////
// C L A S S
///////////////////////////////////////////////////////////////////////////////

class Service extends Daemon
{
    ///////////////////////////////////////////////////////////////////////////////
    // C O N S T A N T S
    ///////////////////////////////////////////////////////////////////////////////

    // Config
    const BLOCKPAGE_PORT      = '6373';

    const CONF_DIR            = '/etc/adamnet';
    const WEBSERVER_CONFIG    = '/etc/httpd/conf/httpd.conf';

    // Defaults
    const BRAIN_FQDN    = 'controller.adamnet.works';
    const LOG_FILE      = '/var/log/anmgr.log';

    // Key paths
    const PATH_PK_FILE        = '/etc/adamnet/signature.pk';
    const PATH_SK_FILE        = '/etc/adamnet/signature.sk';
    const PATH_CKG            = '/usr/local/sbin/anckg';

    ///////////////////////////////////////////////////////////////////////////////
    // V A R I A B L E S
    ///////////////////////////////////////////////////////////////////////////////

    protected $issues = array();
    protected $proxy_supported = NULL;
    protected $warnings = array();

    // Initialized by constructor
    protected $configs = [];

    ///////////////////////////////////////////////////////////////////////////////
    // M E T H O D S
    ///////////////////////////////////////////////////////////////////////////////

    /**
     * Service constructor.
     */
    public function __construct()
    {
        clearos_profile(__METHOD__, __LINE__);

        parent::__construct('anmgr');

        $this->configs = [
            'app' => [self::CONF_DIR . '/app.conf'],
            'mgr' => [self::CONF_DIR . '/anmgr.conf'],
            'dnsmasq' => ['/etc/dnsmasq.d/app-adamone.conf'],
            'dnsmasq_sys' => ['/etc/dnsmasq.conf', self::CONF_DIR . '/dnsmasq-changed.conf'],
            'sysctl' => ['/usr/lib/sysctl.d/10-anmgr.conf'],  # Run after 00-system
        ];
    }

    /**
     * Setup all config files
     *
     * @return void
     */
    public function setup()
    {
        clearos_profile(__METHOD__, __LINE__);

        // Create config folder
        $folder = new Folder(self::CONF_DIR);
        if (!$folder->exists()) {
            $folder->create('root', 'root', '0755');
        }

        // Set up
        $this->_create_keys();

        // Upgrade configurations (if needed)
        $this->_upgrade();

        // Set boxid
        $config = $this->_config('app');
        if (!$config->get('boxid')) {
            $box_id = $this->get_box_id();
            if ($box_id > 0) {
                $config->set('boxid', $box_id);
            } else {
                $this->issues[] = 'Unable to find LAN interface.';
            }
        }

        return !$this->issues;
    }

    /**
     * Verify settings are valid.
     *
     * @return array error messages
     */
    public function verify_setup()
    {
        clearos_profile(__METHOD__, __LINE__);

        $this->update_config();

        // check firewall settings
        $issue = $this->update_firewall();
        $this->add_issue($issue);

        // if needed, restart daemons to pick up any config changes
        if ($this->is_enabled() && count($this->get_issues())) {
            $this->restart_daemons();
        }

        // update the daemons
        $this->update_daemons();
    }

    /**
     * Add an issue to the list.
     *
     * @param string $issue Issue description
     *
     * @return void
     */
    public function add_issue($issue)
    {
        clearos_profile(__METHOD__, __LINE__);

        if ($issue) {
            $this->issues[] = $issue;
        }
    }

    /**
     * Add warning to the list.
     *
     * @param string $message Warning detail
     * @param string $title Warning summary
     *
     * @return void
     */
    public function add_warning($message)
    {
        clearos_profile(__METHOD__, __LINE__);

        $this->warnings[] = array(
            'message' => $message,
        );
    }

    /**
     * Returns an array of issues that need to be resolved before automatic setup can be completed
     *
     * @return array issues
     */
    public function get_issues()
    {
        clearos_profile(__METHOD__, __LINE__);

        return $this->issues;
    }

    /**
     * Reset list of issues. Used to clear false reports after enabling or disabling.
     *
     * @return void
     */
    public function reset_issues()
    {
        clearos_profile(__METHOD__, __LINE__);

        $this->issues = array();
    }

    /**
     * Returns an array of warnings for the user
     *
     * @return array issues
     */
    public function get_warnings()
    {
        clearos_profile(__METHOD__, __LINE__);
        $warnings = $this->warnings;

        $network = new Network();
        $mode = $network->get_mode();
        $internal_dns = $network->get_internal_dns_ips();

        // Standalone mode
        if ($mode != 'gateway' && $mode != 'trustedgateway') {
            $warnings[] = array(
                'title' => $this->_lang('standalone_title'),
                'message' => $this->_lang('standalone_msg'),
            );
        } else if (!$this->_proxy_supported()) {
            // If not using standalone, warn about conflict with subnet proxy
            $warnings[] = array(
                'title' => $this->_lang('subnet_proxy_not_supported'),
                'message' => $this->_lang('subnet_proxy_conflicts_with_webserver'),
            );
        }

        // Using internal DNS
        if ($internal_dns) {
            $warnings[] = array(
                'title' => $this->_lang('internal_dns_title'),
                'message' => $this->_lang('internal_dns_msg'),
            );
        }

        // Interface warnings
        $lan_ifs = $network->get_interfaces(Network::FILTER_LAN);
        $dhcp_ifs = array();
        foreach ($lan_ifs as $lan_if => $lan_data) {
            if (!isset($lan_data['ifcfg']['ipaddr'])) {
                $dhcp_ifs[] = $lan_if;
            }
        }

        if ($dhcp_ifs) {
            $warnings[] = array(
                'title' => $this->_lang('dhcp_ifs_title'),
                'message' => $this->_lang('dhcp_ifs_msg') . implode(', ', $dhcp_ifs) . '.',
            );
        }

        // Recommend use of subnet proxy (when not in use)
        $http_address = $this->get_http_address();
        if ($this->is_enabled() && $this->_proxy_supported() && strpos($http_address, '/') === FALSE) {
            $warnings[] = array(
                'title' => $this->_lang('subnet_proxy_recommended'),
                'message' => $this->_lang('subnet_proxy_recommended_msg'),
            );
        }

        return $warnings;
    }

    /**
     * Get enabled status.
     *
     * @return bool
     */
    public function is_enabled()
    {
        clearos_profile(__METHOD__, __LINE__);

        $config = $this->_config('app');
        return $config->get('enabled');
    }

    /**
     * Validates an IPv4 network address.
     *
     * @param string $network network address in format 172.27.0.0/16
     *
     * @return bool true if valid
     */

    public static function is_valid_ipv4_network($network)
    {
        clearos_profile(__METHOD__, __LINE__);

        $network_utils = new Network_Utils();
        try {
            return $network_utils->is_valid_ipv4_network($network);
        } catch (Validation_Exception $e) {
            return FALSE;
        }
    }

    /**
     * Get block_private_subnets status.
     *
     * @return bool
     */
    public function get_block_private_subnets()
    {
        clearos_profile(__METHOD__, __LINE__);

        $config = $this->_config('app');
        return ($config->get('block-private-subnets') === 'true');

    }

    /**
     * Set block_private_subnets.
     *
     * @param bool $block_private_subnets Block private subnets setting
     *
     * @return bool updated
     */
    public function set_block_private_subnets($block_private_subnets)
    {
        clearos_profile(__METHOD__, __LINE__);

        $block_private_subnets = ($block_private_subnets) ? 'true' : 'false';

        return $this->_config('app')->set('block-private-subnets', $block_private_subnets);
    }

    /**
     * Set enabled status.
     *
     * @param bool $enabled Enabled setting
     *
     * @return bool updated
     */
    public function set_enabled($enabled)
    {
        clearos_profile(__METHOD__, __LINE__);

        if ($this->is_enabled() === $enabled)
            return FALSE;

        if ($enabled) {
            $this->enable();
        } else {
            $this->disable();
        }

        $this->reset_issues();

        return TRUE;
    }

    /**
     * Get HTTP listen address.
     *
     * @return string
     */
    public function get_http_address()
    {
        clearos_profile(__METHOD__, __LINE__);

        $config = $this->_config('app');
        $subnet = $config->get('http-proxy-subnet');
        $ip = $config->get('http-listen-address');

        if ($subnet) {
            return $subnet;
        }

        if ($ip) {
            return $ip;
        }

        return NULL;
    }

    /**
     * Set http listen address.
     *
     * @param bool $address IP address or network
     *
     * @return bool updated
     */
    public function set_http_address($address)
    {
        clearos_profile(__METHOD__, __LINE__);

        $config = $this->_config('app');

        $saved_address = $config->get('http-listen-address');
        $saved_subnet = $config->get('http-proxy-subnet');

        $network_utils = new Network_Utils();

        // Determine if this is a subnet.
        $subnet = $this->is_valid_ipv4_network($address) ? $address : FALSE;
        if ($subnet) {

            $network_utils = new Network_Utils();
            if ($network_utils->ip_on_network('172.27.27.27', $subnet)) {
                // Fun default.
                $address = '172.27.27.27';
            } else {
                // Suggest the first available address in subnet.
                $ip = explode('/', $subnet)[0];
                $address = long2ip( ip2long($ip) + 1 );
            }
        }

        // See if this is changed
        if ($saved_address === $address && $saved_subnet == $subnet) {
            return FALSE;
        }

        // Save changes
        $config->set('http-listen-address', $address);
        $config->set('http-proxy-subnet', $subnet);

        return TRUE;
    }

    /**
     * Enable.
     *
     * @return bool
     */
    public function enable()
    {
        clearos_profile(__METHOD__, __LINE__);

        // 1. Set enabled.
        $this->_config('app')->set('enabled', TRUE);

        // 2. Update all conf files. Must be done before restarting daemons.
        $this->update_config();

        // 3. Add firewall rules.
        $this->update_firewall();

        // 4. Restart the daemons. Filtering should start working.
        $this->restart_daemons();

        // 5. Configure daemon startup scripts.
        $this->update_daemons();

    }

    /**
     * Disable.
     *
     * @return bool
     */
    public function disable()
    {
        clearos_profile(__METHOD__, __LINE__);

        // 1. Remove enabled.
        $this->_config('app')->set('enabled', FALSE);

        // 2. Update config.
        $this->update_config();

        // 3. Restart dnsmasq.
        $this->_daemon_cmd('dnsmasq', 'restart');

        // 4. Remove firewall rules.
        $this->update_firewall();

        // 5. Handle the daemons.
        $this->update_daemons();
    }

    /**
     * Update config.
     *
     * @return void
     */
    public function update_config()
    {
        clearos_profile(__METHOD__, __LINE__);

        $this->update_app_config();
        $this->generate_mgr_config();

        if ($this->is_enabled()) {
            $this->setup_system();
        } else {
            $this->cleanup_system();
        }

        $this->configure_interfaces();
    }

    /**
     * Process the app config.
     *
     * @return array
     */
    public function update_app_config() {
        $config = $this->_config('app');

        // Setup defaults for blocking private subnets
        if ($config->get('block-private-subnets') === FALSE) {
            $network = new Network();
            $internal_dns = $network->get_internal_dns_ips();
            $mode = $this->_get_mode();

            if (($mode == 'gateway' || $mode == 'trustedgateway') && !$internal_dns) {
                $block_private_subnets = 'true';
            } else {
                $block_private_subnets = 'false';
            }

            $config->set('block-private-subnets', $block_private_subnets);
        }

        // Set default log file values.
        if ($config->get('log-file') === FALSE) {
            $config->set('log-file', self::LOG_FILE);
            $config->set('log-level', '0');
        }

        // If enabled and no http address, suggest one.
        if ($this->is_enabled()) {
            $http_address = $this->get_http_address();
            $result = $this->validate_http_address($http_address);
            if ($result) {
                $http_address = $this->suggest_http_address();
                if (empty($http_address)) {
                    $this->add_warning($this->_lang('blockpage_ip_suggest_failed'));
                } else {
                    $this->set_http_address($http_address);
                    $msg = $this->_lang('blockpage_ip_autoupdated');
                    $this->add_warning("$msg: $http_address.");
                }
            }
        }

    }

    /**
     * Generate the mgr config.
     *
     * @return array changed keys
     */
    public function generate_mgr_config() {
        // Reset the values.
        $config = $this->_config('mgr');
        $original = $config->get_all();
        $config->clear();

        $config->comment($this->_lang('autogenerated'));

        // Standard options.
        $config->set('daemon', TRUE);
        $config->set('dns-listen-address', '127.0.0.1@5553');
        $config->set('edns0-accept', TRUE);
        $config->set('enable-tuntap', TRUE);
        $config->set('files-to-monitor',
            '/etc/hosts,hosts;'.
            '/var/lib/dnsmasq/dnsmasq.leases,dnsmasq_leases;'.
            '/etc/resolv-peerdns.conf,resolv'
        );
        $config->set('http-listen-port', self::BLOCKPAGE_PORT);
        $config->set('ipe-extra-ttl', '100');
        $config->set('ipe-min-ttl', '60');
        $config->set('ipe-max-ttl', '120');
        $config->set('lock-file', '/var/lock/anmgr.lock');
        $config->set('pid-file', '/var/run/anmgr.pid');
        $config->set('signature-pk-file', self::PATH_PK_FILE);
        $config->set('signature-sk-file', self::PATH_SK_FILE);
        $config->set('ssl-ca-file', '/etc/adamnet/cert.pem');
        $config->set('brain-fqdn', self::BRAIN_FQDN);

        // Load more information.
        $app_config = $this->_config('app');
        $full_app_config = $app_config->get_all();
        $network = new Network();

        // Get and sort LAN names for consistency.
        $lans = $network->get_interfaces(Network::FILTER_LAN);
        $lan_names = array_keys($lans);
        sort($lan_names);
        $lan_names = implode(',', $lan_names);
        $config->set('lan-iface', $lan_names);

        // Bridge mode requires promiscuous packet monitor.
        if ($this->_get_mode() === 'trustedgateway') {
            $config->set('packet-monitor-promiscuous', TRUE);
        }

        //
        if ($app_config->get('block-private-subnets') === 'false') {
            $config->set('disable-private-subnets', TRUE);
        }

        // Set up mybox redirect.
        $ifaces = $network->get_interfaces(Network::FILTER_LAN);
        foreach ($ifaces as $name => $details) {
            $iface = new Iface($name);
        }
        $lan_ip = $iface->get_live_ip();
        if ($lan_ip) {
            $config->set('mybox-redirect', "https://$lan_ip:81");
        }

        // Copy over any modified settings from app config.
        $ignore_keys = ['enabled', 'block-private-subnets', 'version'];
        foreach ($full_app_config as $key => $value) {
            if (in_array($key, $ignore_keys)) {
                continue;
            }

            $config->set($key, $value);
        }

        // Check for changes.
        $changes = $config->get_changed($original);
        if ($changes) {
            $issue = $this->_lang('app_changed');
            $this->add_issue($issue);
        }

        return (bool)$changes;
    }

    /**
     * Configure any needed system settings. Called if enabled.
     *
     * @return array
     */
    public function setup_system() {

        if (!$this->is_enabled()) {
            throw new Engine_Exception("Cannot call setup if service is disabled.", CLEAROS_ERROR);
        }

        $changes = 0;

        // Configure defaults.
        $dnsmasq_config = $this->_config('dnsmasq');
        $dnsmasq_config_original = $dnsmasq_config->get_all();
        $dnsmasq_config->clear();
        $dnsmasq_config->set('add-subnet', '32,128');
        $dnsmasq_config->set('no-poll', TRUE);
        $dnsmasq_config->set('no-resolv', TRUE);
        $dnsmasq_config->set('server', '127.0.0.1#5553');

        if ($dnsmasq_config->get_changed($dnsmasq_config_original)) {
            $changes += 1;
        }

        // Configure overrides.
        $dnsmasq_sys_config = $this->_config('dnsmasq_sys');
        $changes += $dnsmasq_sys_config->set('cache-size', '0');

        // Remove invalid options.
        $changes += $dnsmasq_sys_config->set('bogus-priv', FALSE);

        // Remove duplicate options (specified in app config).
        $changes += $dnsmasq_sys_config->set('no-poll', FALSE);
        $changes += $dnsmasq_sys_config->set('no-resolv', FALSE);

        // Configure bridge settings.
        $bridge_setting = ($this->_get_mode() == 'trustedgateway') ? '1' : FALSE;
        $bridge_config = $this->_config('sysctl');
        $changes += $bridge_config->set('net.bridge.bridge-nf-call-arptables', $bridge_setting);
        $changes += $bridge_config->set('net.bridge.bridge-nf-call-ip6tables', $bridge_setting);
        $changes += $bridge_config->set('net.bridge.bridge-nf-call-iptables', $bridge_setting);

        // Check for changes.
        if ($changes) {
            $issue = $this->_lang('system_changed');
            $this->add_issue($issue);
        }

    }

    /**
     * Clean up any system settings. Called if disabled.
     *
     * @return array
     */
    public function cleanup_system() {
        if ($this->is_enabled()) {
            throw new Engine_Exception("Cannot call cleanup if service is enabled.", CLEAROS_ERROR);
        }

        $dnsmasq_config = $this->_config('dnsmasq');
        $dnsmasq_config->clear();

        $dnsmasq_sys_config = $this->_config('dnsmasq_sys');
        $dnsmasq_sys_config->restore();
    }

    /**
     * Configure subnet proxy interface.
     *
     * @return array
     */
    public function configure_interfaces() {
        // Handle subnet IP's
        $config = $this->_config('mgr');

        // Load the interface
        $iface = new Iface('lo:adamone');
        try {
            $iface_info = $iface->get_info();
        } catch (Not_Found_Exception $e) {
            $iface_info = array(
                'address' => FALSE,
                'netmask' => FALSE,
            );
        }

        $proxy_subnet = $config->get('http-proxy-subnet');
        if ($this->is_enabled() && $proxy_subnet) {
            // Subnet proxy enabled, so set it up.
            list($ip, $prefix) = explode('/', $proxy_subnet);
            $network_utils = new Network_Utils();
            $netmask = $network_utils->get_netmask($prefix);

            // If not configured with current values, we need to create it.
            if ($iface_info['address'] !== $ip || $iface_info['netmask'] != $netmask) {
                $iface->save_virtual_config($ip, $netmask);
                $iface->enable();

                $this->add_issue($this->_lang('subnet_not_configured'));
            }
        } elseif (!empty($iface_info['address'])) {
            // Remove subnet if it is currently configured.
            $iface->delete_config();
        }

    }

    /**
     * Update daemons.
     *
     * @return void
     */
    public function update_daemons()
    {
        clearos_profile(__METHOD__, __LINE__);

        $daemons = array('anmgr');
        foreach ($daemons as $daemon) {
            try {
                $bootstate = $this->_daemon_cmd($daemon, 'get_boot_state');
            } catch (Engine_Exception $e) {
                $this->add_issue($this->_lang('anmgr_boot_not_found'));
                continue;
            }

            if ($this->is_enabled()) {
                // Enable daemon autostart scripts
                if (!$bootstate) {
                    $this->_daemon_cmd($daemon, 'set_boot_state', TRUE);
                    $this->add_issue($this->_lang($daemon.'_boot_not_set'));
                }

                // Start the daemons
                try {
                    $running = $this->_daemon_cmd($daemon, 'get_running_state');
                } catch (Engine_Exception $e) {
                    // First attempt can raise exception if condition changed recently.
                    $running = $this->_daemon_cmd($daemon, 'get_running_state');
                }
                if (!$running) {
                    try {
                        $this->_daemon_cmd($daemon, 'set_running_state', TRUE);
                        $this->add_issue($this->_lang($daemon.'_not_running'));
                    } catch (Engine_Exception $e) {
                        $this->add_issue($this->_lang($daemon.'_launch_failed'));
                    }
                }
            } else {
                // Disable daemon autostart scripts
                if (!$bootstate) {
                    $this->_daemon_cmd($daemon, 'set_boot_state', FALSE);
                }

                // Stop the daemons
                try {
                    $running = $this->_daemon_cmd($daemon, 'get_running_state');
                } catch (Engine_Exception $e) {
                    // First attempt can raise exception if condition changed recently.
                    $running = $this->_daemon_cmd($daemon, 'get_running_state');
                }
                if ($running) {
                    $this->_daemon_cmd($daemon, 'set_running_state', FALSE);
                }
            }
        }
    }

    /**
     * Verify firewall is set correctly.
     *
     * @return string error message
     */
    public function update_firewall()
    {
        clearos_profile(__METHOD__, __LINE__);

        $firewall = new Firewall();
        $changed = $firewall->update($this->is_enabled(), $this->_config('mgr'));
        if ($changed) {
            return $this->_lang('firewall_error');
        }
    }

    /**
     * Returns the mac address in decimal format.
     *
     * @return string mac address in decimal format
     */
    public function get_box_id()
    {
        clearos_profile(__METHOD__, __LINE__);

        $config = $this->_config('app');

        // Usually this should be stored as a config option.
        if ($config->get('boxid')) {
            return $config->get('boxid');
        }

        $network = new Network();
        $ifs = $network->get_interfaces(Network::FILTER_LAN);
        $if = NULL;
        foreach($ifs as $key => $value) {
            $if = $value;
            break;
        }

        // Make sure we have an interface with a MAC address.
        if ($if === NULL || !$if['hwaddress']) {
            return 0;
        }

        // convert to decimal
        $clean_mac = str_replace(':', '', $if['hwaddress']);
        $box_id = number_format(hexdec($clean_mac), 0, '', '');

        return $box_id;
    }

    /**
     * Restart daemons to reload config.
     *
     * @return void
     */
    public function restart_daemons()
    {
        clearos_profile(__METHOD__, __LINE__);

        $daemons = array('dnsmasq', 'anmgr');

        foreach ($daemons as $daemon) {
            $this->_daemon_cmd($daemon, 'restart');
        }
    }

    /**
     * Suggest an IP or subnet for blockpage.
     *
     * @return string IP
     */
    public function suggest_http_address()
    {
        clearos_profile(__METHOD__, __LINE__);
        if ($this->_proxy_supported()) {
            return $this->_suggest_http_subnet();
        } else {
            return $this->_suggest_http_ip();
        }
    }

    /**
     * Suggest a subnet for block page.
     *
     * @return string IP
     */
    private function _suggest_http_subnet()
    {
        clearos_profile(__METHOD__, __LINE__);

        // use a default
        $address = '172.27.0.0/16';
        if (!$this->validate_http_address($address)) {
            return $address;
        }

        // just find anything
        for ($a = 16; $a <= 32; $a++) {
            $address = "172.$a.0.0/16";
            if (!$this->validate_http_address($address)) {
                return $address;
            }
        }

        for ($a = 0; $a <= 254; $a++) {
            $address = "10.$a.0.0/16";
            if (!$this->validate_http_address($address)) {
                return $address;
            }
        }
    }

    /**
     * Validate block_private_subnets.
     *
     * @param bool $block_private_subnets status
     */
    public function validate_block_private_subnets($block_private_subnets)
    {
        clearos_profile(__METHOD__, __LINE__);
    }

    /**
     * Validate enabled.
     *
     * @param bool $enabled status
     */
    public function validate_enabled($enabled)
    {
        clearos_profile(__METHOD__, __LINE__);
    }

    /**
     * Validate IP or subnet for blockpage.
     *
     * @param string $validate_address Address to validate
     *
     * @return string error message if IP is not valid
     */
    public function validate_http_address($validate_address)
    {
        clearos_profile(__METHOD__, __LINE__);

        // Only check use if it is not already set. If it is set, it's probably bound.
        $check_use = ($validate_address !== $this->get_http_address());
        if (!$check_use) {
            $config = $this->_config('app');
            if ($config->get('http-listen-address') === $validate_address) {
                $check_use = FALSE;
            }
        }

        $network_utils = new Network_Utils();
        if ($this->is_valid_ipv4_network($validate_address)) {
            // Network given
            return $this->_validate_http_subnet($validate_address, $check_use);
        } elseif ($network_utils->is_valid_ip($validate_address)) {
            // IP given
            return $this->_validate_http_ip($validate_address, $check_use);
        } else {
            return $this->_lang('blockpage_ip_invalid_ip_or_network');
        }
    }

    /**
     * Validate network for gateway block page.
     *
     * @param string $validate_network Network address to validate
     *
     * @return string error message if network is not valid
     */
    private function _validate_http_subnet($validate_network, $check_use)
    {
        clearos_profile(__METHOD__, __LINE__);

        // Validate subnet proxy is supported
        if (!$this->_proxy_supported()) {
            return $this->_lang('blockpage_subnet_not_supported');
        }

        // Subnet cannot overlap with any of the interface networks
        $network = new Network();
        $result = $network->network_overlaps_interfaces($validate_network);
        if ($result) {
            return $this->_lang('blockpage_ip_network_overlap');
        }

        // Make sure network is fully in a private range
        $network_info = $network->get_network_info($validate_network);

        if ($network_info['prefix'] < 16) {
            return $this->_lang('blockpage_ip_prefix_too_big');
        }

        $network_utils = new Network_Utils();
        if (
            !$network_utils->is_private_ip($network_info['address']) ||
            !$network_utils->is_private_ip($network_info['broadcast'])
        ) {
            return $this->_lang('blockpage_ip_not_private');
        }

        // Make sure this is not in use by any routes
        $shell = new Shell();
        $shell->execute(Routes::COMMAND_IP, 'route show');
        $output = $shell->get_output();

        if (!empty($output)) {
            foreach ($output as $line) {
                // Get the first word.
                $route_address = explode(' ', $line)[0];

                // If not valid, skip.
                if (!$this->is_valid_ipv4_network($route_address)) {
                    continue;
                }

                $route_info = $network->get_network_info($route_address);
                if (
                    $network_utils->ip_on_network($network_info['address'], $route_address) ||
                    $network_utils->ip_on_network($network_info['broadcast'], $route_address) ||
                    $network_utils->ip_on_network($route_info['address'], $validate_network) ||
                    $network_utils->ip_on_network($route_info['broadcast'], $validate_network)
                ) {
                    return $this->_lang('blockpage_ip_route_overlap');
                }
            }
        }

        // Quick check if network address is in use.
        if ($check_use) {
            if ($network->is_ip_in_use($network_info['address'])) {
                return $this->_lang('blockpage_ip_in_use');
            }
        }
    }

    /**
     * Validate IP for block page.
     *
     * @param string $validate_ip IP address to validate
     *
     * @return string error message if IP is not valid
     */
    private function _validate_http_ip($validate_ip, $check_use)
    {
        clearos_profile(__METHOD__, __LINE__);

        // validate IP
        $resolver = new Resolver();
        $result = $resolver->validate_ip($validate_ip);
        if ($result) {
            return $result;
        }

        if ($this->_get_mode() === 'gateway' || $this->_get_mode() === 'trustedgateway') {
            return $this->_validate_http_ip_gateway($validate_ip, $check_use);
        } else {
            return $this->_validate_http_ip_standalone($validate_ip, $check_use);
        }
    }

    /**
     * Validate IP for block page in gateway mode.
     *
     * @param string $validate_ip IP address to validate
     *
     * @return string error message if IP is not valid
     */
    private function _validate_http_ip_gateway($validate_ip, $check_use)
    {
        clearos_profile(__METHOD__, __LINE__);

        $network = new Network();

        // Validate IP is not on any networks
        if ($network->ip_on_networks($validate_ip)) {
            // If IP is on networks, use standalone validation.
            return $this->_validate_http_ip_standalone($validate_ip, $check_use);
        }

        // Make sure IP address is in a private range
        $network_utils = new Network_Utils();
        if (!$network_utils->is_private_ip($validate_ip)) {
            return $this->_lang('blockpage_ip_not_private');
        }

        // Make sure IP isn't in use.
        if ($check_use) {
            if ($network->is_ip_in_use($validate_ip)) {
                return $this->_lang('blockpage_ip_in_use');
            }
        }
    }

    /**
     * Validate IP for standalone block page.
     *
     * @param string $validate_ip IP address to validate
     *
     * @return string error message if IP is not valid
     */
    private function _validate_http_ip_standalone($validate_ip, $check_use)
    {
        clearos_profile(__METHOD__, __LINE__);

        $network = new Network();
        $lans = $network->get_interfaces(Network::FILTER_LAN);

        // Here's the error
        foreach ($lans as $if_name => $if_info) {

            // validate IP is on subnet
            $iface = new Iface($if_name);
            $network_info = $iface->get_network();

            // If we don't have network info, skip this
            if (empty($network_info)) {
                continue;
            }

            // See if the IP belongs to this network
            $network_utils = new Network_Utils();
            if (!$network_utils->ip_on_network($validate_ip, $network_info)) {
                continue;
            }

            // check that IP is outside DHCP range
            $dnsmasq = new Dnsmasq();
            $range = $dnsmasq->get_subnet($if_name);
            $ip_long = ip2long($validate_ip);
            if ($ip_long >= ip2long($range['start']) && $ip_long <= ip2long($range['end'])) {
                return $this->_lang('blockpage_ip_in_dhcp_range')." ($range[start] - $range[end])";
            }

            // Make sure IP isn't in use.
            if ($check_use) {
                if ($network->is_ip_in_use($validate_ip)) {
                    return $this->_lang('blockpage_ip_in_use');
                }
            }

            // IP is valid
            return;
        }

        // IP wasn't on any of the LAN networks
        return $this->_lang('blockpage_ip_outside_network');
    }


    ///////////////////////////////////////////////////////////////////////////////
    // P R I V A T E   M E T H O D S
    ///////////////////////////////////////////////////////////////////////////////

    /**
     * Return text in proper translation
     *
     * @return void
     */
    public function _lang($key) {
        $util = new Util();
        return $util->lang($key);
    }

    /**
     * Create all the necessary keys.
     *
     * @return void
     */
    protected function _create_keys() {
        $shell = new Shell();

    	if (!file_exists(self::PATH_PK_FILE) || !file_exists(self::PATH_SK_FILE)) {
    		echo "Creating key pair.\n";
            $shell->execute(self::PATH_CKG, self::PATH_PK_FILE . ' ' . self::PATH_SK_FILE);
    	}
    }

    /**
     * Upgrade from old versions.
     *
     * @return void
     */
    protected function _upgrade()
    {
        $app_config = $this->_config('app');
        $mgr_config = $this->_config('mgr');

        $config_v = intval($app_config->get('version', TRUE));

        // --- 2018? --- //
        // Versions that used redirect-local were always enabled, so mark this
        //   new version enabled.
        if ($mgr_config->get('redirect-local')) {
            $app_config->set('enabled', TRUE);
        }

        // --- 2019-11 --- //
        // Mgr config is now auto-generated, so move settings to app config.
        if ($config_v < 1) {
            $keys = array('boxid', 'http-proxy-subnet', 'http-listen-address', 'log-level', 'log-file');
            foreach ($keys as $key) {
                $value = $mgr_config->get($key);
                if ($value && !$app_config->get($key)) {
                    $app_config->set($key, $value);
                }
            }
        }

        $app_config->set('version', 1);
    }

    /**
     * Get network mode.
     *
     * @return string mode
     */
    private function _get_mode()
    {
        clearos_profile(__METHOD__, __LINE__);

        $network = new Network();
        return $network->get_mode();
    }

    /**
     * Check if the web server installed.
     *
     * @return bool true if installed
     */
    private function _proxy_supported()
    {
        clearos_profile(__METHOD__, __LINE__);

        if (is_null($this->proxy_supported)) {

            if ($this->_get_mode() !== 'gateway' && $this->_get_mode() !== 'trustedgateway') {
                $this->proxy_supported = FALSE;
            } else {
                // Proxy is not supported if webserver is installed.
                $this->proxy_supported = !file_exists(self::WEBSERVER_CONFIG);
            }
        }

        return $this->proxy_supported;
    }

    /**
     * return proper config file.
     *
     * @param string $which Which config file
     *
     * @return  string config file location
     * @throws Engine_Exception
     */
    protected function _config($which)
    {
        clearos_profile(__METHOD__, __LINE__);

        if (!isset($this->configs[$which])) {
            throw new Engine_Exception("Invalid config type: '$which'.", CLEAROS_ERROR);
        }
        $paths = $this->configs[$which];
        if (!isset($paths[1])) {
            $paths[] = NULL;
        }

        return new Configurator($paths[0], $paths[1]);
    }

    /**
     * Setup a hosts entry.
     *
     * @return bool
     */
    protected function _setup_host($host, $ip, $aliases = array())
    {
        $hosts = new Hosts();

        // remove invalid IPs for these hosts
        $all_hosts = array_merge(array($host), $aliases);
        foreach ($all_hosts as $current_host) {
            while (
                $hosts->get_ip_by_hostname($current_host)
                && $hosts->get_ip_by_hostname($current_host) != $ip
            ) {
                $hosts->delete_entry($hosts->get_ip_by_hostname($current_host));
            }
        }

        // setup host (if needed)
        foreach ($all_hosts as $current_host) {
            if (!$hosts->get_ip_by_hostname($current_host)) {
                if ($hosts->entry_exists($ip)) {
                    $hosts->edit_entry($ip, $host, $aliases);
                } else {
                    $hosts->add_entry($ip, $host, $aliases);
                }
                return TRUE;
            }
        }

        return FALSE;
    }

    /**
     * Suggest an IP for block page.
     *
     * @return string IP
     */
    private function _suggest_http_ip()
    {
        clearos_profile(__METHOD__, __LINE__);

        // If valid listen address already exists, see if we can use that.
        $config = $this->_config('app');
        $listen_address = $config->get('http-listen-address');
        if ($listen_address) {
            if ($this->validate_http_address($listen_address)) {
                return $listen_address;
            }
        }

        if ($this->_get_mode() === 'gateway' || $this->_get_mode() === 'trustedgateway') {
            // Gateway mode can use any IP address.

            // use a default
            $ip = '172.27.27.27';
            if (!$this->validate_http_address($ip))
                return $ip;

            // just find anything
            for ($a = 16; $a <= 32; $a++) {
                $ip = "172.$a.$a.$a";
                if (!$this->validate_http_address($ip)) {
                    return $ip;
                }
            }

        }

        // Get an IP on the network.
        $network = new Network();
        $lan_ifs = $network->get_interfaces(Network::FILTER_LAN);
        $lan_if = NULL;
        foreach ($lan_ifs as $key => $value) {
            $lan_if = $value;
            break;
        }

        if ($lan_if === NULL) {
            return NULL;
        }

        // get iface info
        $network_utils = new Network_Utils();
        $address = $network_utils->get_network_address($lan_if['address'], $lan_if['netmask']);
        $address = ip2long($address) + 1;
        $broadcast = ip2long($lan_if['broadcast']);

        // find an acceptable IP
        while ($address < $broadcast) {
            // validate current address
            $ip = long2ip($address);
            if (!$this->validate_http_address($ip)) {
                // found acceptable address
                return $ip;
            }

            // check next address
            $address += 1;
        }
    }

    /**
     * Runs the specified daemon command.
     *
     * @param string $script  Name
     * @param string $command Command
     * @param string $value Value (only needed for some items)
     *
     * @return string/bool
     * @throws Engine_Exception
     */
    protected function _daemon_cmd($script, $command, $value=FALSE)
    {
        clearos_profile(__METHOD__, __LINE__);

        $daemon = new Daemon($script);

        switch ($command) {
            case 'get_boot_state':
                return $daemon->get_boot_state();
            case 'set_boot_state':
                return $daemon->set_boot_state($value);
            case 'restart':
                return $daemon->restart();
            case 'get_status':
                return $daemon->get_status(); // text
            case 'get_running_state':
                return $daemon->get_running_state(); // code
            case 'set_running_state':
                return $daemon->set_running_state($value);
            default:
                throw new Engine_Exception("Invalid command: '$command'.", CLEAROS_ERROR);
        }
    }

}
